package org

import java.util.Date
import java.util.Properties

import scala.Iterable

import org.apache.spark.SparkConf
import org.apache.spark.storage.StorageLevel
import org.apache.spark.streaming.Seconds
import org.apache.spark.streaming.State
import org.apache.spark.streaming.StateSpec
import org.apache.spark.streaming.StreamingContext
import org.apache.spark.streaming.dstream.DStream
import org.apache.spark.streaming.dstream.DStream.toPairDStreamFunctions
import org.apache.spark.streaming.dstream.MapWithStateDStream
import org.infinispan.spark.stream.InfinispanDStream
import org.infinispan.spark.stream.InfinispanInputDStream

/**
 * <p>Computes average sensor data in each place from incoming stream of measurements.</p>
 * <p>The measurement stream is read from the JDG's sensor-data and produces stream of average data.
 * Updates are written back to the Data Grid into a different cache called sensor-avg-data.
 */
object Analyzer {

  val inputDataGridCache = "sensor-data"
  val outputDataGridCache = "sensor-avg-data"


  
  val mapFunc = (place: String, currDataLists: Option[Iterable[ShipmentData]], state: State[Map[String, Iterable[ShipmentData]]]) => {

    // obtain the state map contains the past 5 minutes shipment data
    val stateMap: Map[String, Iterable[ShipmentData]] = state.getOption().getOrElse(Map[String, Iterable[ShipmentData]]())
    val pastData: (Iterable[ShipmentData]) = stateMap.getOrElse(place, (Iterable[ShipmentData])())

    var sumTemp: Double = 0d;
    var sumHumd: Double = 0d;
    val currTime: Long = java.lang.System.currentTimeMillis();
    var skuNos: String = "";

    //create the new list for storing the data, from both pastData and currData
    var newShipmentData = collection.mutable.ListBuffer[ShipmentData]();

    //only count the time in less than 5 minutes
    pastData.foreach { data: ShipmentData =>
      val sensorTime: Long = data.getSensorTime();

      //for testing usage      
      val sensorTimeInDisplay = new Date(sensorTime);

      //only include the data from past 5 minutes
      if (currTime - sensorTime < 5 * 60 * 1000) {
        sumTemp = sumTemp + data.getTemperature();
        sumHumd = sumHumd + data.getHumidity();
        newShipmentData += data;
      }
    }

    val currData: Iterable[ShipmentData] = currDataLists.getOrElse(List());

    currData.foreach { data: ShipmentData =>
      //only count the time in less than 5 minutes
      val sensorTime: Long = data.getSensorTime();

      //for testing usage      
      val sensorTimeInDisplay = new Date(sensorTime);
      //only include the data from past 5 minutes
      if (currTime - sensorTime < 5 * 60 * 1000) {
        sumTemp = sumTemp + data.getTemperature();
        sumHumd = sumHumd + data.getHumidity();
        newShipmentData += data;
        skuNos = data.getSkuNumbers();
      }
    }

    val count = newShipmentData.size;
     var avgTemp: Double = sumTemp / count;
    var avgHumd: Double = sumHumd / count;

    var avgShipmentData: ShipmentData = new ShipmentData();
    avgShipmentData.setTemperature(avgTemp);
    avgShipmentData.setHumidity(avgHumd);
    avgShipmentData.setSkuNumbers(skuNos);
    avgShipmentData.setSensorTime(currTime);

    // update stored state
    state.update(stateMap.updated(place, newShipmentData.toList))

    (place, avgShipmentData)
  }

  
  def main(args: Array[String]) {

    val usage = """
    Please input the Spark server address in ip:port format
  """

    if (args.length == 0) println(usage)
    val dataGridServer = args(0)
    // Initialize the Spark streaming context, with a batch duration of 1 second.
    val sparkConf = new SparkConf().setAppName("SparkSensor")
    val ssc = new StreamingContext(sparkConf, Seconds(1))
    // set up checkpoint to store state information
    ssc.checkpoint("/mnt/shared/sharefs/spark-sensor")

    // configure the connection to the DataGrid and create the incoming DStream
    val configIn = new Properties
    configIn.put("infinispan.rdd.cacheName", inputDataGridCache)
    configIn.put("infinispan.client.hotrod.server_list", dataGridServer)
    val ispnStream = new InfinispanInputDStream[String, ShipmentData](ssc, StorageLevel.MEMORY_ONLY, configIn)

    // extract the (place, temperature) pair from the incoming stream that is composed of (key, value, eventType)
    // we are assuming only events of type CLIENT_CACHE_ENTRY_CREATED will be present.
    val measurementStream: DStream[(String, ShipmentData)] = ispnStream map { case (key, value, eventType) => (key, value) }

    // as measurements are batched, it can contain several measurements from the same place - let's group them together
    val measurementGrouped: DStream[(String, Iterable[ShipmentData])] = measurementStream.groupByKey()

    // Produce a new DStream combining the previous DStream with the stored state
    val avgSensorDataStream: MapWithStateDStream[String, Iterable[ShipmentData], Map[String, Iterable[ShipmentData]], (String, ShipmentData)] =
      measurementGrouped.mapWithState(StateSpec.function(mapFunc))

    // just print number of items in each RDD
    avgSensorDataStream.foreachRDD(rdd => {
      printf("# items in DStream: %d\n", rdd.count())
      rdd.foreach { case (place, average) => println("Averages:" + place + " -> " + average) }
    })

    // write stream to the avg-sensor-data cache
    val configOut = new Properties
    configOut.put("infinispan.rdd.cacheName", outputDataGridCache)
    configOut.put("infinispan.client.hotrod.server_list", dataGridServer)
    avgSensorDataStream.writeToInfinispan(configOut)

    ssc.start()
    ssc.awaitTermination()

  }
}

